@@ -52,11 +52,13 @@ typedef struct {
   int last_char_idx;
   double last_pos_y;
   double last_pos_x;
+  double scale_x;
+  double scale_y;
 } paps_private_t;
 
 
 // Forward declarations
-static void add_postscript_prologue(GString *ps_string);
+static void add_postscript_prologue(paps_private_t *paps);
 static gchar *get_next_char_id_strdup(paps_private_t *paps);
 static void add_line_to_postscript(paps_private_t *paps,
 				   GString *line_str,
@@ -76,12 +78,30 @@ paps_t *paps_new()
     paps->last_pos_x = -1e67;
     paps->last_pos_y = -1e67;
     paps->last_char_idx = 0;
+    paps->scale_x = 1.0;
+    paps->scale_y = 1.0;
 
-    add_postscript_prologue(paps->header);
+    add_postscript_prologue(paps);
 
     return paps;
 }
 
+void
+paps_set_scale(paps_t  *paps_,
+	       gdouble  scale_x,
+	       gdouble  scale_y)
+{
+    paps_private_t *paps = (paps_private_t *)paps_;
+
+    paps->scale_x = scale_x;
+    paps->scale_y = scale_y;
+#if 0
+    // Why would you want to do such a thing???
+    g_string_erase(paps->header, 0, -1);
+    add_postscript_prologue(paps);
+#endif
+}
+
 PangoContext *paps_get_pango_context()
 {
   return pango_ft2_get_context (PAPS_DPI, PAPS_DPI);
@@ -114,6 +134,7 @@ typedef struct {
 /* Information passed in user data when drawing outlines */
 typedef struct _OutlineInfo OutlineInfo;
 struct _OutlineInfo {
+  paps_private_t *paps;
   GString *out_string;
   FT_Vector glyph_origin;
   int dpi;
@@ -221,9 +242,10 @@ gchar *paps_layout_line_to_postscript_strdup(paps_t *paps_,
   return ret_str;
 }
 
-void add_postscript_prologue(GString *ps_string)
+static void
+add_postscript_prologue(paps_private_t *paps)
 {
-  g_string_append_printf(ps_string,
+  g_string_append_printf(paps->header,
 			 "%%%%BeginProlog\n"
 			 "/papsdict 1 dict def\n"
 			 "papsdict begin\n"
@@ -231,7 +253,7 @@ void add_postscript_prologue(GString *ps_string)
 			 );
   
   /* Outline support */
-  g_string_append_printf(ps_string,
+  g_string_append_printf(paps->header,
 			 "/conicto {\n"
 			 "    /to_y exch def\n"
 			 "    /to_x exch def\n"
@@ -268,7 +290,7 @@ void add_postscript_prologue(GString *ps_string)
   // a packed version of the pango layout data. Currently it just executes
   // the symbols corresponding to the encoded characters, but in the future
   // it will also contain some meta data, e.g. the size of the layout.
-  g_string_append_printf(ps_string,
+  g_string_append_printf(paps->header,
 			 "/paps_exec {\n"
 			 "  1 dict begin\n"
 			 "  /ps exch def\n"
@@ -320,7 +342,7 @@ void add_postscript_prologue(GString *ps_string)
 			 );
 
   /* Open up dictionaries */
-  g_string_append(ps_string,
+  g_string_append(paps->header,
 		  "/fontdict 1 dict def\n"
 		  "papsdict begin fontdict begin\n");
 }
@@ -389,7 +411,7 @@ static void draw_contour(paps_private_t *paps,
           glyph_pos_x = x_pos + 1.0*geometry.x_offset * scale;
           glyph_pos_y = line_start_pos_y - 1.0*geometry.y_offset * scale;
 
-	  x_pos += geometry.width * scale;
+	  x_pos += geometry.width * scale * paps->scale_x;
 
           if (glyphs->glyphs[glyph_idx].glyph == PANGO_GLYPH_EMPTY)
             continue;
@@ -399,8 +421,7 @@ static void draw_contour(paps_private_t *paps,
 			      ft_face,
 			      &glyphs->glyphs[glyph_idx],
 			      glyph_pos_x,
-			      glyph_pos_y
-			      );
+			      glyph_pos_y);
         }
 
       runs_list = runs_list->next;
@@ -415,13 +436,12 @@ void draw_bezier_outline(paps_private_t *paps,
                          FT_Face face,
                          PangoGlyphInfo *glyph_info,
                          double pos_x,
-                         double pos_y
-                         )
+                         double pos_y)
 {
   static gchar glyph_hash_string[100];
   double scale = 72.0 / PANGO_SCALE  / PAPS_DPI;
   double epsilon = 1e-2;
-  double glyph_width = glyph_info->geometry.width * scale;
+  double glyph_width = glyph_info->geometry.width * scale * paps->scale_x;
   gchar *id = NULL;
 
   /* Output outline */
@@ -464,6 +484,7 @@ void draw_bezier_outline(paps_private_t *paps,
           
           // Create the outline
           outlinefunc = &ps_outlinefunc;
+          outline_info.paps = paps;
           outline_info.glyph_origin.x = pos_x;
           outline_info.is_empty = 1;
           outline_info.glyph_origin.y = pos_y;
@@ -480,7 +501,7 @@ void draw_bezier_outline(paps_private_t *paps,
           g_string_append_printf(glyph_def_string,
                                  "%.0f fwd_x\n"
                                  "end_ol\n",
-                                 glyph_info->geometry.width * scale * PAPS_DPI
+                                 glyph_info->geometry.width * scale * paps->scale_x * PAPS_DPI
                                  );
           
           // TBD - Check if the glyph_def_string is empty. If so, set the
@@ -488,7 +509,7 @@ void draw_bezier_outline(paps_private_t *paps,
           if (outline_info.is_empty
               || glyph_info->glyph == 0) 
             id[0] = '*';
-          else 
+          else
             // Put the font in the font def dictionary
             g_string_append_printf(paps->header,
                                    "/%s { %s } def\n",
@@ -559,8 +580,8 @@ static int paps_ps_move_to( FT_Vector* to,
   OutlineInfo *outline_info = (OutlineInfo*)user_data;
   g_string_append_printf(outline_info->out_string,
                          "%d %d m\n",
-                         (int)(to->x * FT2PS) ,
-                         (int)(to->y * FT2PS ));
+                         (int)(to->x * FT2PS*outline_info->paps->scale_x) ,
+                         (int)(to->y * FT2PS*outline_info->paps->scale_y ));
   return 0;
 }
 
@@ -570,8 +591,8 @@ static int paps_ps_line_to( FT_Vector*  to,
   OutlineInfo *outline_info = (OutlineInfo*)user_data;
   g_string_append_printf(outline_info->out_string,
                          "%d %d l\n",
-                         (int)(to->x * FT2PS) ,
-                         (int)(to->y * FT2PS) );
+                         (int)(to->x * FT2PS * outline_info->paps->scale_x) ,
+                         (int)(to->y * FT2PS * outline_info->paps->scale_y) );
   outline_info->is_empty = 0;
   return 0;
 }
@@ -583,10 +604,10 @@ static int paps_ps_conic_to( FT_Vector*  control,
   OutlineInfo *outline_info = (OutlineInfo*)user_data;
   g_string_append_printf(outline_info->out_string,
                          "%d %d %d %d x\n",
-                         (int)(control->x * FT2PS)  ,
-                         (int)(control->y * FT2PS) ,
-                         (int)(to->x * FT2PS),
-                         (int)(to->y * FT2PS));
+                         (int)(control->x * FT2PS*outline_info->paps->scale_x)  ,
+                         (int)(control->y * FT2PS*outline_info->paps->scale_y) ,
+                         (int)(to->x * FT2PS*outline_info->paps->scale_x),
+                         (int)(to->y * FT2PS*outline_info->paps->scale_y));
   outline_info->is_empty = 0;
   return 0;
 }
@@ -599,12 +620,12 @@ static int paps_ps_cubic_to( FT_Vector*  control1,
   OutlineInfo *outline_info = (OutlineInfo*)user_data;
   g_string_append_printf(outline_info->out_string,
                          "%d %d %d %d %d %d c\n",
-                         (int)(control1->x * FT2PS) , 
-                         (int)(control1->y * FT2PS) ,
-                         (int)(control2->x * FT2PS) ,
-                         (int)(control2->y * FT2PS) ,
-                         (int)(to->x * FT2PS) ,
-                         (int)(to->y * FT2PS) );
+                         (int)(control1->x * FT2PS*outline_info->paps->scale_x) , 
+                         (int)(control1->y * FT2PS*outline_info->paps->scale_y) ,
+                         (int)(control2->x * FT2PS*outline_info->paps->scale_x) ,
+                         (int)(control2->y * FT2PS*outline_info->paps->scale_y) ,
+                         (int)(to->x * FT2PS*outline_info->paps->scale_x) ,
+                         (int)(to->y * FT2PS*outline_info->paps->scale_y) );
   outline_info->is_empty = 0;
   return 0;
 }
