@@ -22,15 +22,21 @@
 
 
 #include <pango/pango.h>
+#include <pango/pangoft2.h>
 #include "libpaps.h"
 #include <errno.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
+#include <locale.h>
 
 #define BUFSIZE 1024
-#define HEADER_FONT_SCALE 12
+#define DEFAULT_FONT_FAMILY	"Monospace"
+#define DEFAULT_FONT_SIZE	"12"
+#define HEADER_FONT_FAMILY	"Monospace Bold"
+#define HEADER_FONT_SCALE	"12"
+#define MAKE_FONT_NAME(f,s)	f " " s
 
 typedef enum {
     PAPER_TYPE_A4 = 0,
@@ -66,6 +72,8 @@ typedef struct {
   int header_sep;
   int header_height;
   int footer_height;
+  gdouble scale_x;
+  gdouble scale_y;
   gboolean do_draw_header;
   gboolean do_draw_footer;
   gboolean do_duplex;
@@ -74,9 +82,14 @@ typedef struct {
   gboolean do_justify;
   gboolean do_separation_line;
   gboolean do_draw_contour;
+  gboolean do_wordwrap;
+  gboolean do_use_markup;
+  gboolean do_stretch_chars;
   PangoDirection pango_dir;
   gchar *filename;
   gchar *header_font_desc;
+  gint lpi;
+  gint cpi;
 } page_layout_t;
 
 typedef struct {
@@ -88,6 +101,7 @@ typedef struct {
   PangoLayoutLine *pango_line;
   PangoRectangle logical_rect;
   PangoRectangle ink_rect;
+  int formfeed;
 } LineLink;
 
 typedef struct _Paragraph Paragraph;
@@ -98,12 +112,15 @@ struct _Paragraph {
   char *text;
   int length;
   int height;   /* Height, in pixels */
+  int formfeed;
   PangoLayout *layout;
 };
 
 /* Information passed in user data when drawing outlines */
-GList        *split_paragraphs_into_lines  (GList           *paragraphs);
-static char  *read_file                    (FILE            *file);
+GList        *split_paragraphs_into_lines  (page_layout_t   *page_layout,
+					    GList           *paragraphs);
+static char  *read_file                    (FILE            *file,
+                                            GIConv           handle);
 static GList *split_text_into_paragraphs   (PangoContext    *pango_context,
                                             page_layout_t   *page_layout,
                                             int              paint_width,
@@ -181,18 +198,99 @@ _paps_arg_paper_cb(const char *option_name,
   return retval;
 }
 
+static gboolean
+_paps_arg_lpi_cb(const gchar *option_name,
+		 const gchar *value,
+		 gpointer     data)
+{
+  gboolean retval = TRUE;
+  gchar *p = NULL;
+  page_layout_t *page_layout = (page_layout_t*)data;
+
+  if (value && *value)
+    {
+      errno = 0;
+      page_layout->lpi = g_strtod(value, &p);
+      if ((p && *p) || errno == ERANGE)
+        {
+          fprintf(stderr, "given LPI value was invalid.\n");
+          retval = FALSE;
+        }
+    }
+  else
+    {
+      fprintf(stderr, "You must specify the amount of lines per inch.\n");
+      retval = FALSE;
+    }
+
+  return retval;
+}
+
+static gboolean
+_paps_arg_cpi_cb(const gchar *option_name,
+		 const gchar *value,
+		 gpointer     data)
+{
+  gboolean retval = TRUE;
+  gchar *p = NULL;
+  page_layout_t *page_layout = (page_layout_t*)data;
+  
+  if (value && *value)
+    {
+      errno = 0;
+      page_layout->cpi = g_strtod(value, &p);
+      if ((p && *p) || errno == ERANGE)
+        {
+          fprintf(stderr, "given CPI value was invalid.\n");
+          retval = FALSE;
+        }
+    }
+  else
+    {
+      fprintf(stderr, "You must specify the amount of characters per inch.\n");
+      retval = FALSE;
+    }
+
+  return retval;
+}
+
+static PangoLanguage *
+get_language(void)
+{
+  PangoLanguage *retval;
+  gchar *lang = g_strdup (setlocale (LC_CTYPE, NULL));
+  gchar *p;
+
+  p = strchr (lang, '.');
+  if (p)
+    *p = 0;
+  p = strchr (lang, '@');
+  if (p)
+    *p = 0;
+
+  retval = pango_language_from_string (lang);
+  g_free (lang);
+
+  return retval;
+}
+
 int main(int argc, char *argv[])
 {
   gboolean do_landscape = FALSE, do_rtl = FALSE, do_justify = FALSE, do_draw_header = FALSE;
-  int num_columns = 1, font_scale = 12;
+  gboolean do_stretch_chars = FALSE;
+  gboolean do_use_markup = FALSE;
+  gboolean do_wordwrap = TRUE; // What should be default?
+  int num_columns = 1;
   int top_margin = 36, bottom_margin = 36, right_margin = 36, left_margin = 36;
-  char *font_family = "Monospace";
+  gchar *font = MAKE_FONT_NAME (DEFAULT_FONT_FAMILY, DEFAULT_FONT_SIZE), *encoding = NULL;
+  page_layout_t page_layout;
   GOptionContext *ctxt = g_option_context_new("[text file]");
   GOptionEntry entries[] = {
     {"landscape", 0, 0, G_OPTION_ARG_NONE, &do_landscape, "Landscape output. (Default: portrait)", NULL},
+    {"stretch-chars", 0, 0, G_OPTION_ARG_NONE, &do_stretch_chars, "Whether to stretch characters in y-direction to fill lines. (Default: no)", NULL},
+    {"markup", 0, 0, G_OPTION_ARG_NONE, &do_use_markup, "Should the text be considered pango markup? (Default: no)", NULL},
     {"columns", 0, 0, G_OPTION_ARG_INT, &num_columns, "Number of columns output. (Default: 1)", "NUM"},
-    {"font_scale", 0, 0, G_OPTION_ARG_INT, &font_scale, "Font scaling. (Default: 12)", "NUM"},
-    {"family", 0, 0, G_OPTION_ARG_STRING, &font_family, "Pango FT2 font family. (Default: Monospace)", "FAMILY"},
+    {"font", 0, 0, G_OPTION_ARG_STRING, &font, "Set the font description. (Default: Monospace 12)", "DESC"},
     {"rtl", 0, 0, G_OPTION_ARG_NONE, &do_rtl, "Do rtl layout.", NULL},
     {"justify", 0, 0, G_OPTION_ARG_NONE, &do_justify, "Do justify the lines.", NULL},
     {"paper", 0, 0, G_OPTION_ARG_CALLBACK, _paps_arg_paper_cb,
@@ -203,19 +301,22 @@ int main(int argc, char *argv[])
     {"right-margin", 0, 0, G_OPTION_ARG_INT, &right_margin, "Set right margin. (Default: 36)", "NUM"},
     {"left-margin", 0, 0, G_OPTION_ARG_INT, &left_margin, "Set left margin. (Default: 36)", "NUM"},
     {"header", 0, 0, G_OPTION_ARG_NONE, &do_draw_header, "Draw page header for each page.", NULL},
+    {"encoding", 0, 0, G_OPTION_ARG_STRING, &encoding, "Assume the documentation encoding.", "ENCODING"},
+    {"lpi", 0, 0, G_OPTION_ARG_CALLBACK, _paps_arg_lpi_cb, "Set the amount of lines per inch.", "REAL"},
+    {"cpi", 0, 0, G_OPTION_ARG_CALLBACK, _paps_arg_cpi_cb, "Set the amount of characters per inch.", "REAL"},
     {NULL}
+
   };
   GError *error = NULL;
-  char *filename_in;
-  char *title;
   FILE *IN, *OUT = NULL;
-  page_layout_t page_layout;
-  char *text;
   GList *paragraphs;
   GList *pango_lines;
   PangoContext *pango_context;
   PangoFontDescription *font_description;
   PangoDirection pango_dir = PANGO_DIRECTION_LTR;
+  PangoFontMap *fontmap;
+  PangoFontset *fontset;
+  PangoFontMetrics *metrics;
   int num_pages = 1;
   int gutter_width = 40;
   int total_gutter_width;
@@ -224,13 +325,25 @@ int main(int argc, char *argv[])
   int do_tumble = -1;   /* -1 means not initialized */
   int do_duplex = -1;
   gchar *paps_header = NULL;
-  gchar *header_font_desc = "Monospace Bold 12";
+  gchar *header_font_desc = MAKE_FONT_NAME (HEADER_FONT_FAMILY, HEADER_FONT_SCALE);
+  gchar *filename_in, *title, *text;
   int header_sep = 20;
+  int max_width = 0, w;
+  GIConv cvh = NULL;
+  GOptionGroup *options;
 
   /* Prerequisite when using glib. */
   g_type_init();
-  
+
+  /* Init page_layout_t parameters set by the option parsing */
+  page_layout.cpi = page_layout.lpi = 0;
+
+  options = g_option_group_new("main","","",&page_layout, NULL);
+  g_option_group_add_entries(options, entries);
+  g_option_context_set_main_group(ctxt, options);
+#if 0
   g_option_context_add_main_entries(ctxt, entries, NULL);
+#endif
   
   /* Parse command line */
   if (!g_option_context_parse(ctxt, &argc, &argv, &error))
@@ -263,16 +376,15 @@ int main(int argc, char *argv[])
   pango_context = paps_get_pango_context (paps);
   
   /* Setup pango */
-  pango_context_set_language (pango_context, pango_language_from_string ("en_US"));
+  pango_context_set_language (pango_context, get_language ());
   pango_context_set_base_dir (pango_context, pango_dir);
   
-  font_description = pango_font_description_new ();
-  pango_font_description_set_family (font_description, g_strdup(font_family));
-  pango_font_description_set_style (font_description, PANGO_STYLE_NORMAL);
-  pango_font_description_set_variant (font_description, PANGO_VARIANT_NORMAL);
-  pango_font_description_set_weight (font_description, PANGO_WEIGHT_NORMAL);
-  pango_font_description_set_stretch (font_description, PANGO_STRETCH_NORMAL);
-  pango_font_description_set_size (font_description, font_scale * PANGO_SCALE);
+  /* create the font description */
+  font_description = pango_font_description_from_string (font);
+  if ((pango_font_description_get_set_fields (font_description) & PANGO_FONT_MASK_FAMILY) == 0)
+    pango_font_description_set_family (font_description, DEFAULT_FONT_FAMILY);
+  if ((pango_font_description_get_set_fields (font_description) & PANGO_FONT_MASK_SIZE) == 0)
+    pango_font_description_set_size (font_description, atoi(DEFAULT_FONT_SIZE) * PANGO_SCALE);
 
   pango_context_set_font_description (pango_context, font_description);
 
@@ -314,6 +426,9 @@ int main(int argc, char *argv[])
   page_layout.header_ypos = page_layout.top_margin;
   page_layout.header_height = 0;
   page_layout.footer_height = 0;
+  page_layout.do_wordwrap = do_wordwrap;
+  page_layout.scale_x = 1.0L;
+  page_layout.scale_y = 1.0L;
   if (do_draw_header)
     page_layout.header_sep =  header_sep;
   else
@@ -331,22 +446,68 @@ int main(int argc, char *argv[])
   page_layout.do_separation_line = TRUE;
   page_layout.do_landscape = do_landscape;
   page_layout.do_justify = do_justify;
+  page_layout.do_stretch_chars = do_stretch_chars;
+  page_layout.do_use_markup = do_use_markup;
   page_layout.do_tumble = do_tumble;
   page_layout.do_duplex = do_duplex;
   page_layout.pango_dir = pango_dir;
   page_layout.filename = filename_in;
   page_layout.header_font_desc = header_font_desc;
-  
-  text = read_file(IN);
+
+  /* calculate x-coordinate scale */
+  if (page_layout.cpi > 0.0L)
+    {
+      double scale;
+      fontmap = pango_ft2_font_map_new ();
+      fontset = pango_font_map_load_fontset (fontmap, pango_context, font_description, get_language ());
+      metrics = pango_fontset_get_metrics (fontset);
+      max_width = pango_font_metrics_get_approximate_char_width (metrics);
+      w = pango_font_metrics_get_approximate_digit_width (metrics);
+      if (w > max_width)
+	  max_width = w;
+      page_layout.scale_x = 1 / page_layout.cpi * 72.0 * PANGO_SCALE / max_width;
+      pango_font_metrics_unref (metrics);
+      g_object_unref (G_OBJECT (fontmap));
+
+      // Now figure out how to scale the font to get that size
+      scale = 1 / page_layout.cpi * 72.0 * PANGO_SCALE / max_width;
+
+      fprintf(stderr, "scale = %f\n", scale);
+      
+      // update the font size to that width
+      pango_font_description_set_size (font_description, (int)(atoi(DEFAULT_FONT_SIZE) * PANGO_SCALE * scale));
+      pango_context_set_font_description (pango_context, font_description);
+      
+    }
+
+  page_layout.scale_x = page_layout.scale_y = 1.0;
+      
+
+  if (encoding != NULL)
+    {
+      cvh = g_iconv_open ("UTF-8", encoding);
+      if (cvh == NULL)
+        {
+          fprintf(stderr, "%s: Invalid encoding: %s\n", g_get_prgname (), encoding);
+          exit(-1);
+        }
+    }
+
+  text = read_file(IN, cvh);
+
+  if (encoding != NULL && cvh != NULL)
+    g_iconv_close(cvh);
+
   paragraphs = split_text_into_paragraphs(pango_context,
                                           &page_layout,
                                           page_layout.column_width * page_layout.pt_to_pixel,
                                           text);
-  pango_lines = split_paragraphs_into_lines(paragraphs);
-  
+  pango_lines = split_paragraphs_into_lines(&page_layout, paragraphs);
+
   if (OUT == NULL)
     OUT = stdout;
 
+  paps_set_scale(paps, page_layout.scale_x, page_layout.scale_y);
   print_postscript_header(OUT, title, &page_layout);
   ps_pages_string = g_string_new("");
   
@@ -371,7 +532,8 @@ int main(int argc, char *argv[])
 /* Read an entire file into a string
  */
 static char *
-read_file (FILE *file)
+read_file (FILE   *file,
+           GIConv  handle)
 {
   GString *inbuf;
   char *text;
@@ -380,7 +542,9 @@ read_file (FILE *file)
   inbuf = g_string_new (NULL);
   while (1)
     {
-      char *bp = fgets (buffer, BUFSIZE-1, file);
+      char *ib, *ob, obuffer[BUFSIZE * 6], *bp = fgets (buffer, BUFSIZE-1, file);
+      gsize iblen, oblen;
+
       if (ferror (file))
         {
           fprintf(stderr, "%s: Error reading file.\n", g_get_prgname ());
@@ -390,7 +554,20 @@ read_file (FILE *file)
       else if (bp == NULL)
         break;
 
-      g_string_append (inbuf, buffer);
+      if (handle != NULL)
+        {
+          ib = bp;
+          iblen = strlen (ib);
+          ob = bp = obuffer;
+          oblen = BUFSIZE * 6 - 1;
+          if (g_iconv (handle, &ib, &iblen, &ob, &oblen) == -1)
+            {
+              fprintf (stderr, "%s: Error while converting strings.\n", g_get_prgname ());
+              return NULL;
+            }
+          obuffer[BUFSIZE * 6 - 1 - oblen] = 0;
+        }
+      g_string_append (inbuf, bp);
     }
 
   fclose (file);
@@ -405,7 +582,11 @@ read_file (FILE *file)
   return text;
 }
 
-/* Take a UTF8 string and break it into paragraphs on \n characters
+#if 0
+/* Take a UTF8 string and break it into paragraphs on \n characters.
+ *
+ * Sorry. I couldn't figure out what this version was supposed to do
+ *
  */
 static GList *
 split_text_into_paragraphs (PangoContext *pango_context,
@@ -419,7 +600,7 @@ split_text_into_paragraphs (PangoContext *pango_context,
   GList *result = NULL;
   char *last_para = text;
   
-  while (*p)
+  while (p != NULL && *p)
     {
       wc = g_utf8_get_char (p);
       next = g_utf8_next_char (p);
@@ -428,20 +609,79 @@ split_text_into_paragraphs (PangoContext *pango_context,
           fprintf (stderr, "%s: Invalid character in input\n", g_get_prgname ());
           wc = 0;
         }
-      if (!*p || !wc || wc == '\n')
+      if (!*p || !wc || wc == '\n' || wc == '\f')
         {
           Paragraph *para = g_new (Paragraph, 1);
           para->text = last_para;
           para->length = p - last_para;
           para->layout = pango_layout_new (pango_context);
+
+	  if (cpi > 0.0L && page_layout->do_wordwrap)
+	    {
+	      PangoRectangle ink_rect, logical_rect;
+	      wchar_t *wtext, *wnewtext;
+	      gchar *newtext;
+	      size_t i, len, wwidth = 0, n;
+
+	      wtext = g_utf8_to_ucs4 (para->text, para->length, NULL, NULL, NULL);
+	      if (wtext == NULL)
+	        {
+		  fprintf (stderr, "Failed to convert UTF-8 to UCS-4.\n");
+		  return NULL;
+		}
+
+	      len = g_utf8_strlen (para->text, para->length);
+	      /* the amount of characters to be able to put on the line against CPI */
+	      n = page_layout->column_width / 72.0 * cpi;
+	      if (len > n)
+	        {
+		  wnewtext = g_new (wchar_t, wcslen (wtext) + 1);
+		  if (wnewtext == NULL)
+		    {
+		      fprintf (stderr, "Failed to allocate a memory.\n");
+		      g_free (wtext);
+		      return NULL;
+		    }
+		  for (i = 0; i < len; i++)
+		    {
+		      wwidth += wcwidth (wtext[i]);
+		      if (wwidth > n)
+			  break;
+		      wnewtext[i] = wtext[i];
+		    }
+		  wnewtext[i] = 0L;
+
+		  newtext = g_ucs4_to_utf8 ((const gunichar *)wnewtext, i, NULL, NULL, NULL);
+		  if (newtext == NULL)
+		    {
+		      fprintf (stderr, "Failed to convert UCS-4 to UTF-8.\n");
+		      return NULL;
+		    }
+
+		  pango_layout_set_text (para->layout, newtext, -1);
+		  pango_layout_get_extents (para->layout, &ink_rect, &logical_rect);
+		  /* update paint_width to wrap_against CPI */
+		  paint_width = logical_rect.width / PANGO_SCALE;
+		  g_free (newtext);
+		  g_free (wnewtext);
+		}
+	      g_free (wtext);
+	    }
           pango_layout_set_text (para->layout, para->text, para->length);
           pango_layout_set_justify (para->layout, page_layout->do_justify);
           pango_layout_set_alignment (para->layout,
                                       page_layout->pango_dir == PANGO_DIRECTION_LTR
                                       ? PANGO_ALIGN_LEFT : PANGO_ALIGN_RIGHT);
           pango_layout_set_width (para->layout, paint_width * PANGO_SCALE);
+          if (page_layout->do_wordwrap)
+              pango_layout_set_wrap (para->layout, PANGO_WRAP_WORD_CHAR);
           para->height = 0;
 
+          if (wc == '\f')
+              para->formfeed = 1;
+          else
+              para->formfeed = 0;
+
           last_para = next;
             
           result = g_list_prepend (result, para);
@@ -453,13 +693,103 @@ split_text_into_paragraphs (PangoContext *pango_context,
 
   return g_list_reverse (result);
 }
+#endif
+
+/* Take a UTF8 string and break it into paragraphs on \n characters
+ */
+static GList *
+split_text_into_paragraphs (PangoContext *pango_context,
+                            page_layout_t *page_layout,
+                            int paint_width,  /* In pixels */
+                            char *text)
+{
+  char *p = text;
+  char *next;
+  gunichar wc;
+  GList *result = NULL;
+  char *last_para = text;
+
+  /* If we are using markup we treat the entire text as a single paragraph.
+   * I tested it and found that this is much slower than the split and
+   * assign method used below. Otherwise we might as well use this single
+   * chunk method always.
+   */
+  if (page_layout->do_use_markup)
+    {
+      Paragraph *para = g_new (Paragraph, 1);
+      para->text = text;
+      para->length = strlen(text);
+      para->layout = pango_layout_new (pango_context);
+      //          pango_layout_set_font_description (para->layout, font_description);
+      pango_layout_set_markup (para->layout, para->text, para->length);
+      pango_layout_set_justify (para->layout, page_layout->do_justify);
+      pango_layout_set_alignment (para->layout,
+                                  page_layout->pango_dir == PANGO_DIRECTION_LTR
+                                      ? PANGO_ALIGN_LEFT : PANGO_ALIGN_RIGHT);
+      pango_layout_set_wrap (para->layout, PANGO_WRAP_WORD_CHAR);
+      
+      pango_layout_set_width (para->layout, paint_width * PANGO_SCALE);
+      para->height = 0;
+      
+      result = g_list_prepend (result, para);
+    }
+  else
+    {
+
+      while (p != NULL && *p)
+        {
+          wc = g_utf8_get_char (p);
+          next = g_utf8_next_char (p);
+          if (wc == (gunichar)-1)
+            {
+              fprintf (stderr, "%s: Invalid character in input\n", g_get_prgname ());
+              wc = 0;
+            }
+          if (!*p || !wc || wc == '\n' || wc == '\f')
+            {
+              Paragraph *para = g_new (Paragraph, 1);
+              para->text = last_para;
+              para->length = p - last_para;
+              para->layout = pango_layout_new (pango_context);
+              //          pango_layout_set_font_description (para->layout, font_description);
+              pango_layout_set_text (para->layout, para->text, para->length);
+              pango_layout_set_justify (para->layout, page_layout->do_justify);
+              pango_layout_set_alignment (para->layout,
+                                          page_layout->pango_dir == PANGO_DIRECTION_LTR
+                                          ? PANGO_ALIGN_LEFT : PANGO_ALIGN_RIGHT);
+              pango_layout_set_wrap (para->layout, PANGO_WRAP_WORD_CHAR);
+
+              pango_layout_set_width (para->layout, paint_width * PANGO_SCALE);
+              para->height = 0;
+
+              last_para = next;
+            
+              if (wc == '\f')
+                para->formfeed = 1;
+              else
+                para->formfeed = 0;
+
+              result = g_list_prepend (result, para);
+            }
+          if (!wc) /* incomplete character at end */
+            break;
+          p = next;
+        }
+    }
+
+  return g_list_reverse (result);
+}
+
+
 
 /* Split a list of paragraphs into a list of lines.
  */
 GList *
-split_paragraphs_into_lines(GList *paragraphs)
+split_paragraphs_into_lines(page_layout_t *page_layout,
+			    GList         *paragraphs)
 {
   GList *line_list = NULL;
+  int max_height = 0;
   /* Read the file */
 
   /* Now split all the pagraphs into lines */
@@ -469,6 +799,7 @@ split_paragraphs_into_lines(GList *paragraphs)
   while(par_list)
     {
       int para_num_lines, i;
+      LineLink *line_link;
       Paragraph *para = par_list->data;
 
       para_num_lines = pango_layout_get_line_count(para->layout);
@@ -476,18 +807,26 @@ split_paragraphs_into_lines(GList *paragraphs)
       for (i=0; i<para_num_lines; i++)
         {
           PangoRectangle logical_rect, ink_rect;
-          LineLink *line_link = g_new(LineLink, 1);
           
+          line_link = g_new(LineLink, 1);
+          line_link->formfeed = 0;
           line_link->pango_line = pango_layout_get_line(para->layout, i);
           pango_layout_line_get_extents(line_link->pango_line,
                                         &ink_rect, &logical_rect);
           line_link->logical_rect = logical_rect;
+          if (para->formfeed && i == (para_num_lines - 1))
+              line_link->formfeed = 1;
           line_link->ink_rect = ink_rect;
           line_list = g_list_prepend(line_list, line_link);
+	  if (logical_rect.height > max_height)
+	      max_height = logical_rect.height;
         }
 
       par_list = par_list->next;
     }
+  
+  if (page_layout->do_stretch_chars && page_layout->lpi > 0.0L)
+      page_layout->scale_y = 1.0 / page_layout->lpi * 72.0 * page_layout->pt_to_pixel * PANGO_SCALE / max_height;
 
   return g_list_reverse(line_list);
   
@@ -504,6 +843,7 @@ output_pages(FILE          *OUT,
   int column_y_pos = 0;
   int page_idx = 1;
   int pango_column_height = page_layout->column_height * page_layout->pt_to_pixel * PANGO_SCALE;
+  LineLink *prev_line_link = NULL;
 
   start_page(OUT, page_idx);
 
@@ -516,8 +856,9 @@ output_pages(FILE          *OUT,
       PangoLayoutLine *line = line_link->pango_line;
       
       /* Check if we need to move to next column */
-      if (column_y_pos + line_link->logical_rect.height
-          >= pango_column_height)
+      if ((column_y_pos + line_link->logical_rect.height
+           >= pango_column_height) ||
+          (prev_line_link && prev_line_link->formfeed))
         {
           column_idx++;
           column_y_pos = 0;
@@ -544,9 +885,14 @@ output_pages(FILE          *OUT,
                         column_y_pos+line_link->logical_rect.height,
                         page_layout,
                         line);
-      column_y_pos += line_link->logical_rect.height;
+
+      if (page_layout->lpi > 0.0L)
+        column_y_pos += (int)(1.0 / page_layout->lpi * 72.0 * page_layout->pt_to_pixel * PANGO_SCALE);
+      else
+        column_y_pos += line_link->logical_rect.height;
       
       pango_lines = pango_lines->next;
+      prev_line_link = line_link;
     }
   eject_page(OUT);
   return page_idx;
@@ -574,7 +920,7 @@ void print_postscript_header(FILE *OUT,
   fprintf(OUT,
           "%%!PS-Adobe-3.0\n"
           "%%%%Title: %s\n"
-          "%%%%Creator: paps version 0.6.3 by Dov Grobgeld\n"
+          "%%%%Creator: paps version 0.6.7 by Dov Grobgeld\n"
           "%%%%Pages: (atend)\n"
           "%%%%BoundingBox: 0 0 %d %d\n"
           "%%%%BeginProlog\n"
@@ -696,6 +1042,8 @@ void print_postscript_header(FILE *OUT,
           "    papsdict begin\n"
           "    gsave\n"
           "    do_landscape {turnpage} if \n"
+          "    %% ps2pdf gets wrong orientation without this!\n"
+          "    /Helvetica findfont setfont 100 100 moveto ( ) show\n"
           "    firstcolumn\n"
           "    end\n"
           "} def\n"
@@ -735,7 +1083,7 @@ void eject_column(FILE *OUT,
   if (column_idx == 1)
     total_gutter = 1.0 * page_layout->gutter_width /2;
   else
-    total_gutter = (column_idx + 1.5) * page_layout->gutter_width;
+    total_gutter = (column_idx - 0.5) * page_layout->gutter_width;
       
   x_pos = page_layout->left_margin
         + page_layout->column_width * column_idx
@@ -802,6 +1150,7 @@ draw_line_to_page(FILE *OUT,
       x_pos += page_layout->column_width  - logical_rect.width / (page_layout->pt_to_pixel * PANGO_SCALE);
   }
 
+  paps_set_scale(paps, page_layout->scale_x, page_layout->scale_y);
   ps_layout = paps_layout_line_to_postscript_strdup(paps,
                                                     x_pos, y_pos,
                                                     line);
@@ -863,6 +1212,7 @@ draw_page_header_line_to_page(FILE            *OUT,
       y_pos = page_layout->page_height - page_layout->top_margin - height;
       page_layout->header_height = height;
     }
+  paps_set_scale(paps, page_layout->scale_x, page_layout->scale_y);
   ps_layout = paps_layout_line_to_postscript_strdup(paps,
                                                     x_pos, y_pos,
                                                     line);
@@ -876,6 +1226,7 @@ draw_page_header_line_to_page(FILE            *OUT,
                                 &ink_rect,
                                 &logical_rect);
   x_pos = (page_layout->page_width - (logical_rect.width / PANGO_SCALE * page_layout->pixel_to_pt)) / 2;
+  paps_set_scale(paps, page_layout->scale_x, page_layout->scale_y);
   ps_layout = paps_layout_line_to_postscript_strdup(paps,
                                                     x_pos, y_pos,
                                                     line);
@@ -889,6 +1240,7 @@ draw_page_header_line_to_page(FILE            *OUT,
                                 &ink_rect,
                                 &logical_rect);
   x_pos = page_layout->page_width - page_layout->right_margin - (logical_rect.width / PANGO_SCALE * page_layout->pixel_to_pt);
+  paps_set_scale(paps, page_layout->scale_x, page_layout->scale_y);
   ps_layout = paps_layout_line_to_postscript_strdup(paps,
                                                     x_pos, y_pos,
                                                     line);
@@ -907,4 +1259,3 @@ draw_page_header_line_to_page(FILE            *OUT,
 
   return logical_rect.height;
 }
-
